module Page = struct
  (* The code in this module is pure, without side effects. *)
  type t = {
    uri : string;
    title : string;
    history : string list;
    id : LablgtkReact.id;
  }

  type init = unit

  type action_functional = [
    | `Uri of string
    | `Title of string
  ]

  type action_ui = [ action_functional | `None ]

  let create =
    let id_gen = LablgtkReact.id_gen () in
    fun () -> {
      uri = "about:blank";
      title = "(empty)";
      history = [];
      id = id_gen ();
    }

  let on_action_func t = function
    | `Uri uri -> { t with uri; history = uri :: t.history }, `Uri uri
    | `Title title -> { t with title }, `Title title

  let new_ () =
   `End, `Before, create ()
end

module Z = struct
  module Z = Zipper
  type t = Page.t Z.t
  type init = unit

  type action_common = [
    | `Add of (Z.where_add * Z.where_remove * Page.t)
    | `Nth of int
    | `RemoveIndex of int
  ]

  type action_functional = [
    action_common
    | `RemoveMe of LablgtkReact.id
    | `Set of Page.t
  ]

  type action_ui = [ action_common | `None ]

  let create () =
    Z.empty

  let on_action_func z (action : action_functional) = match action with
  | `Add (where_add, _, e) as msg-> Z.add where_add z e, msg
  | `Nth i as msg -> Z.nth z i, msg
  | `RemoveMe id ->
      (* TODO: some computations are done several times; check with the
       * implementation of the remove_nth function. *)
      let pos_e = Z.pos (Z.find (fun z -> id = (Z.get z).Page.id) z) in
      Z.remove_nth z pos_e, `RemoveIndex pos_e
  | `RemoveIndex i as msg -> Z.remove_nth z i, msg
  | `Set e -> Z.set z e, `None
end

module X = LablgtkReact.Core (MiniFRP) (Page)
module Set = LablgtkReact.Core (MiniFRP) (Z)

class webview ~link page =
  (* This class uses side-effects and GLib/GTK+ functions. *)
  let core = new X.core () in
  let sink ~webview = function
    | `Uri uri -> webview#load_uri uri
    | `Title title -> print_endline title
    | `None -> ()
  in
  object (self)
  val box = GPack.vbox ()
  val webview = GWebView.web_view () 
  val address = GEdit.entry ()
  method label = new Label.label ~title:page.Page.title ()
  method coerce = box#coerce
  method webview = webview
  method destroy = box#destroy
  initializer
    box#pack ~expand:false address#coerce;
    box#pack ~expand:true webview#coerce;
    ignore (webview#connect#notify_title ~callback:self#label#text#set_text);
    core#connect (sink ~webview);
    ignore (address#connect#activate ~callback:(fun () -> core#send (`Uri address#text)))
end

let gtk_create () =
  let window = GWindow.window ~width:800 ~height:600 ~allow_shrink:true () in
  ignore (window#connect#destroy ~callback:GMain.Main.quit);
  window

let shortcuts =
  let open GtkSignal in
  let create_new_page =
    { name = "create-new-page"; classe = `notebook; marshaller = marshal_unit },
    "<Control>T";
  in
  let close_page =
    { name = "close-page"; classe = `notebook; marshaller = marshal_unit },
    "<Control>W";
  in
  [ create_new_page; close_page ]

let set_shortcuts ~tabs ~core =
  Shortcuts.create_shortcuts shortcuts tabs#as_notebook;
  StdLabels.List.iter shortcuts ~f:(fun (sgn, _shortcut) ->
    let f ~sgn callback =
      ignore (GtkSignal.connect ~sgn ~callback tabs#as_notebook)
    in
    match sgn.GtkSignal.name with
    | "create-new-page" -> f ~sgn (fun () -> core#send (`Add (Page.new_ ())))
    | "close-page" -> f ~sgn (fun () -> core#send (`RemoveIndex tabs#current_page))
    | _ -> ())

let () =
  let window = gtk_create () in
  let core = new Set.core () in
  let link e = core#send (`Set e) in
  let tabs = Tabs.tabs ~packing:window#add ~core ~nouveau:(new webview ~link) in
  set_shortcuts ~tabs ~core;
  core#send (`Add (Page.new_ ()));
  window#show ();
  GtkThread.main ()

